C===================================================================
C
      PROGRAM LABECON
C
C  This program solves the two dimensional (LA)place equation
C  using the (B)oundary (E)lement method with (CON)stant
C  boundary elements
C
      IMPLICIT REAL*8 (A-H,O-Z)
      CHARACTER*15 INPUTFILE,OUTPUTFILE
C
C  Set the maximum dimensions
C
      PARAMETER (N=16)
      PARAMETER (IN=9)
C
C   N= Number of boundary elements which is equal to the number 
C      of boundary nodes
C  IN= Number of internal points where the function u is calculated
C
      DIMENSION INDEX(N)
      DIMENSION XL(N+1),YL(N+1),XM(N),YM(N),G(N,N),H(N,N)
      DIMENSION UB(N),A(N,N),UNB(N),XIN(IN),YIN(IN),UIN(IN)
C
C  Read the names and open the input and output files
C
      WRITE (*,'(A)')' Name of the INPUTFILE (max.15 characters)'
      READ  (*,'(A)') INPUTFILE
      WRITE (*,'(A)')' Name of the OUTPUTFILE (max.15 characters)'
      READ  (*,'(A)') OUTPUTFILE
      OPEN (UNIT=1, FILE=INPUTFILE)
      OPEN (UNIT=2, FILE=OUTPUTFILE)
C
C  Read the data from INPUTFILE
C
      CALL INPUT (XL,YL,XIN,YIN,INDEX,UB,N,IN)
C
C  Compute the G matrix
C
      CALL GMATR (XL,YL,XM,YM,G,N) 
C
C  Compute the H matrix
C
      CALL HMATR (XL,YL,XM,YM,H,N)
C
C  Form the system of equations AX=B
C
      CALL ABMATR (G,H,A,UNB,UB,INDEX,N)
C
C  Solve the system of equations
C
      CALL SOLVEQ(A,UNB,N,LSING)
C
C  Form the vectors U and UN of all the boundary values
C
      CALL REORDER (UB,UNB,INDEX,N)
C
C  Compute the values UIN of u at the internal points
C
      CALL UINTER (XL,YL,XIN,YIN,UB,UNB,UIN,N,IN)
C
C  Print the results in the OUTPUTFILE
C
      CALL OUTPUT (XM,YM,UB,UNB,XIN,YIN,UIN,N,IN)
C
C  Close input and output files
C
      CLOSE(1)
      CLOSE(2)
C
      STOP
      END
C
C
C====================================================================
C
      SUBROUTINE INPUT (XL,YL,XIN,YIN,INDEX,UB,N,IN)
C
C  This subroutine reads the data from the input file
C  and writes them in the output file
C
      IMPLICIT REAL*8 (A-H,O-Z)
      CHARACTER*80 NAME,TITLE
      DIMENSION XL(N+1),YL(N+1),XIN(IN),YIN(IN),INDEX(N),UB(N)
C
      WRITE(2,100)
  100 FORMAT('',69('*'))
C
C  Read user's name
C
      READ(1,'(A)')NAME
C
      WRITE(2,'(A)')NAME
C
C  Read the title of the program
C
      READ(1,'(A)')TITLE
C
      WRITE(2,'(A)')TITLE
C
      WRITE(2,200)N,IN
  200 FORMAT(//'BASIC PARAMETERS'//2X,'NUMBER OF BOUNDARY ELEMENTS='
     1,I3/2X,'NUMBER OF INTERNAL POINTS WHERE THE FUNCTION IS CALCULATED
     1=',I3)
C
C  Read the coordinates XL,YL of the extreme points of the boundary elements
C
      READ(1,*) (XL(I),YL(I),I=1,N)
C
C  Write the coordinates in the output file
C
      WRITE(2,300)
  300 FORMAT(//2X,'COORDINATES OF THE EXTREME POINTS OF THE BOUNDARY '
     1'ELEMENTS',//2X,'POINT',9X,'XL',15X,'YL')
      DO 20 I=1,N
   20 WRITE(2,400) I,XL(I),YL(I)
  400 FORMAT(2X,I3,2(3X,E14.5))
C
C  Read the boundary conditions and store them in UB(I). 
C  If INDEX(I)=0, the value UB(I) is the prescribed value of u. 
C  If INDEX(I)=1, the normal derivative un is prescribed.
C
      READ(1,*) (INDEX(I),UB(I),I=1,N)
C
C  Write the boundary conditions in the output file
C
      WRITE(2,500)
  500 FORMAT(//2X,'BOUNDARY CONDITIONS'//2X,'NODE',6X,'INDEX',
     1 7X,'PRESCRIBED VALUE')
      DO 30 I=1,N
   30 WRITE(2,600) I,INDEX(I),UB(I)
  600 FORMAT (2X,I3,9X,I1,8X,E14.5)
C
C  Read the coordinates of the internal points
C
      READ(1,*) (XIN(I),YIN(I),I=1,IN)
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE GMATR (XL,YL,XM,YM,G,N)
C
C  This subroutine computes the elements of the G matrix
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XL(N+1),YL(N+1),XM(N),YM(N)
      DIMENSION G(N,N)
C
C  Compute the nodal coordinates and store them in the arrays 
C  XM and YM
C
      XL(N+1)=XL(1)
      YL(N+1)=YL(1)
      DO 10 I=1,N
      XM(I)=(XL(I)+XL(I+1))/2.
   10 YM(I)=(YL(I)+YL(I+1))/2.
C
C  Compute the elements of matrix G
C
      DO 20 I=1,N
      DO 20 J=1,N
      JP1=J+1
      IF(I.NE.J)THEN
      CALL RLINTC(XM(I),YM(I),XL(J),YL(J),XL(JP1),YL(JP1),RESULT)
      G(I,J)=RESULT
      ELSEIF(I.EQ.J)THEN
      CALL SLINTC(XL(J),YL(J),XL(JP1),YL(JP1),RESULT)
      G(I,J)=RESULT
      ENDIF
   20 CONTINUE
      RETURN
      END
C
C
C================================================================
C
      SUBROUTINE RLINTC(X0,Y0,X1,Y1,X2,Y2,RESULT)
C
C  This subroutine computes the off-diagonal elements of the 
C  matrix G 
C
C  RA= The distance of point O from the Gauss integration point 
C      on the boundary element
C
C  WG= The weights of the Gauss integration
C
C  XI= The coordinates of the Gauss integration points in the 
C      interval [-1,1]
C
C  XC,YC= The global coordinates of the Gauss integration points
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XC(4),YC(4),XI(4),WG(4)
      DATA XI/-0.86113631,-0.33998104,0.33998104,0.86113631/
      DATA WG/0.34785485,0.65214515,0.65214515,0.34785485/
      PI=ACOS(-1.)
      AX=(X2-X1)/2.
      AY=(Y2-Y1)/2.
      BX=(X2+X1)/2.
      BY=(Y2+Y1)/2.
C
C  Compute the line integral
C
      RESULT=0.
C
      DO 30 I=1,4
      XC(I)=AX*XI(I)+BX
      YC(I)=AY*XI(I)+BY
      RA=SQRT((XC(I)-X0)**2+(YC(I)-Y0)**2)
   30 RESULT=RESULT+DLOG(RA)*WG(I)
      SL=2.*SQRT(AX**2+AY**2)
      RESULT=RESULT*SL/(4.*PI)
      RETURN
      END
C
C
C=================================================================
C
      SUBROUTINE SLINTC(X1,Y1,X2,Y2,RESULT)
C
C  This subroutines computes the diagonal elements of the matrix G
C
      IMPLICIT REAL*8 (A-H,O-Z)
      PI=ACOS(-1.)
      AX=(X2-X1)/2.
      AY=(Y2-Y1)/2.
      SL=SQRT(AX**2+AY**2)
      RESULT=SL*(DLOG(SL)-1.)/PI
      RETURN
      END
C
C
C=================================================================
C
      SUBROUTINE HMATR(XL,YL,XM,YM,H,N)
C
C  This subroutine computes the elements of the H matrix
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XL(N+1),YL(N+1),XM(N),YM(N)
      DIMENSION H(N,N)
C
      PI=ACOS(-1.)
C
      XL(N+1)=XL(1)
      YL(N+1)=YL(1)
      DO 10  I=1,N
      XM(I)=(XL(I)+XL(I+1))/2.
   10 YM(I)=(YL(I)+YL(I+1))/2.
C
C  Compute the elements of the H matrix
C
      DO 20  I=1,N
      DO 20  J=1,N
      IF (I.NE.J) THEN      
      CALL DALPHA (XM(I),YM(I),XL(J),YL(J),XL(J+1),YL(J+1),RESULT)
      H(I,J)=RESULT
      ELSEIF (I.EQ.J) THEN
      H(I,J)=-0.5
      ENDIF
   20 CONTINUE
      RETURN
      END
C
C
C=================================================================
C
      SUBROUTINE DALPHA(X0,Y0,X1,Y1,X2,Y2,RESULT)
C
C  This subroutine computes the off diagonal elements of 
C  the H matrix
C
      IMPLICIT REAL*8 (A-H,O-Z)
      PI=ACOS(-1.)
      DY1=Y1-Y0
      DX1=X1-X0
      DY2=Y2-Y0
      DX2=X2-X0
      DL1=SQRT(DX1**2+DY1**2)
      COS1=DX1/DL1
      SIN1=DY1/DL1
      DX2R=DX2*COS1+DY2*SIN1
      DY2R=-DX2*SIN1+DY2*COS1
      DA=ATAN2(DY2R,DX2R)
      RESULT=DA/(2.*PI)
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE ABMATR (G,H,A,UNB,UB,INDEX,N)
C
C  This subroutine rearranges the matrices G and H and produces 
C  the matrices A and B=UNB
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION G(N,N),H(N,N),A(N,N),UNB(N),UB(N),INDEX(N)
C
C  Reorder the columns of the system of equations and 
C  store them in A
C
      DO 40 J=1,N
      IF(INDEX(J).EQ.0) THEN
      DO 20 I=1,N
   20 A(I,J)=-G(I,J)
      ELSEIF(INDEX(J).NE.0) THEN
      DO 30 I=1,N
   30 A(I,J)=H(I,J)
      END IF
   40 CONTINUE
C
C  Compute the right hand side vector and store it in UNB
C
      DO 50 I=1,N
      UNB(I)=0.
      DO 60 J=1,N
      IF(INDEX(J).EQ.0) THEN
      UNB(I)=UNB(I)-H(I,J)*UB(J) 
      ELSEIF(INDEX(J).NE.0)THEN
      UNB(I)=UNB(I)+G(I,J)*UB(J)
      ENDIF
   60 CONTINUE
   50 CONTINUE
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE SOLVEQ(A,UNB,N,LSING)
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(N,N),UNB(N)
      CALL LEQS(A,UNB,N,LSING)
      IF(LSING.EQ.0)THEN
      WRITE(2,150)
  150 FORMAT(/,'',69('*')//2X'The system has been solved regularly'/)
      ELSEIF(LSING.EQ.1)THEN
      WRITE(2,170)
  170 FORMAT(/,'',69('*')//2X'The system is singular'/)
      ENDIF
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE LEQS(A,B,N,LSING)
C
C   This subroutine uses Gauss elimination to solve
C   a system of linear equations, [A]{X}={B}, where
C   A : One-dimensional array which contains the occasional row of
C       the two-dimensional array of the coefficients of the unknowns
C   B : One-dimensional array which contains the known coefficients
C   N : Integer denoting the number of the unknowns
C   LSING: Integer taking the values:
C         LSING = 0, if the system has been solved regularly
C         LSING = 1, if the system is singular
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION A(1),B(1)
C
      LSING=0
      JJ=-N
      DO 10 J=1,N
        JY=J+1
        JJ=JJ+N+1
        AMAX=0.0
        IHELP=JJ-J
      DO 20 I=J,N
        IJ=IHELP+I
        IF(ABS(AMAX)-ABS(A(IJ))) 30,20,20
   30      AMAX=A(IJ)
           IMAX=I
   20 CONTINUE
         IF(ABS(AMAX).EQ.0.) THEN
            LSING=1
            RETURN
         END IF
      I1=J+N*(J-2)
      IHELP=IMAX-J
      DO 40 K=J,N
        I1=I1+N
        I2=I1+IHELP
        ATEMP=A(I1)
        A(I1)=A(I2)
        A(I2)=ATEMP
   40 A(I1)=A(I1)/AMAX
      ATEMP=B(IMAX)
      B(IMAX)=B(J)
      B(J)=ATEMP/AMAX
         IF(J-N) 50,70,50
   50       IQS=N*(J-1)
            DO 10 IX=JY,N
              IXJ=IQS+IX
              IHELP=J-IX
              DO 60 JX=JY,N
                IJREF=N*(JX-1)+IX
                JJX=IJREF+IHELP
   60           A(IJREF)=A(IJREF)-(A(IXJ)*A(JJX))
   10 B(IX)=B(IX)-B(J)*A(IXJ)
   70    NY=N-1
      NN=N*N
      DO 80 J=1,NY
        I1=NN-J
        I2=N-J
        I3=N
      DO 80 K=1,J
        B(I2)=B(I2)-A(I1)*B(I3)
        I1=I1-N
   80   I3=I3-1
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE REORDER(UB,UNB,INDEX,N)
C
C  This subroutine rearranges the arrays UB and UNB in such a way
C  that all values of the function u are stored in UB while all
C  the values of the normal derivative un are stored in UNB
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION UB(N),UNB(N),INDEX(N)
C
      DO 20  I=1,N
      IF(INDEX(I))20,20,10
   10 CH=UB(I)
      UB(I)=UNB(I)
      UNB(I)=CH
   20 CONTINUE
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE UINTER(XL,YL,XIN,YIN,UB,UNB,UIN,N,IN)
C
C  This subroutine computes the values of u at the internal points
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XL(N+1),YL(N+1),XIN(IN),YIN(IN),UB(N),UNB(N),UIN(IN)
C
C  Compute the values of u at the internal points
C
      DO 10  K=1,IN
      UIN(K)=0.
      DO 20  J=1,N
      JP1=J+1
      CALL DALPHA(XIN(K),YIN(K),XL(J),YL(J),XL(JP1),YL(JP1),RESH)
      CALL RLINTC(XIN(K),YIN(K),XL(J),YL(J),XL(JP1),YL(JP1),RESG)
   20 UIN(K)=UIN(K)+RESH*UB(J)-RESG*UNB(J)
   10 CONTINUE
      RETURN
      END
C
C
C==================================================================
C
      SUBROUTINE OUTPUT(XM,YM,UB,UNB,XIN,YIN,UIN,N,IN)
C
C  This subroutine prints the results in the output file.
C
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION XM(N),YM(N),UB(N),UNB(N)
      DIMENSION XIN(IN),YIN(IN),UIN(IN)
C
      WRITE(2,100)
  100 FORMAT('',69('*')///1X,'RESULTS'///2X,'BOUNDARY NODES'//
     1 11X,'X',15X,'Y',15X,'U',15X,'Un'/)
C
      DO 10 I=1,N
   10 WRITE(2,200) XM(I),YM(I),UB(I),UNB(I)
  200 FORMAT(4(2X,E14.5))
C
      WRITE (2,300)
  300 FORMAT(//,2X,'INTERNAL POINTS'//10X,'X',15X,'Y',11X,
     1 'SOLUTION U'/)
C
      DO 20 K=1,IN
   20 WRITE(2,400) XIN(K),YIN(K),UIN(K)
  400 FORMAT(3(2X,E14.5))
      WRITE(2,500)
  500 FORMAT(/,'',69('*'))
      RETURN
      END          
C
C
C==================================================================